from __future__ import print_function
import Pyro4


# a Chain member. Passes messages to the next link,
# until the message went full-circle: then it exits.

class Chain(object):
    def __init__(self, name, next_node):
        self.name = name
        self.nextName = next_node
        self.next = None

    @Pyro4.expose
    def process(self, message):
        if self.next is None:
            self.next = Pyro4.core.Proxy("PYRONAME:example.chain." + self.nextName)
        if self.name in message:
            print("Back at %s; we completed the circle!" % self.name)
            return ["complete at " + self.name]
        else:
            print("I'm %s, passing to %s" % (self.name, self.nextName))
            message.append(self.name)
            result = self.next.process(message)
            result.insert(0, "passed on from " + self.name)
            return result



# mandelbrot fractal,  z=z^2+c
from __future__ import print_function, division
import time
try:
    import tkinter
except ImportError:
    import Tkinter as tkinter

from server import MandelbrotColorPixels

res_x = 1000
res_y = 800


class MandelWindow(object):
    def __init__(self):
        self.root = tkinter.Tk()
        self.root.title("Mandelbrot (Single Core)")
        canvas = tkinter.Canvas(self.root, width=res_x, height=res_y, bg="#000000")
        canvas.pack()
        self.img = tkinter.PhotoImage(width=res_x, height=res_y)
        canvas.create_image((res_x/2, res_y/2), image=self.img, state="normal")
        self.mandel = MandelbrotColorPixels()
        self.start_time = time.time()
        self.root.after(1000, lambda: self.draw_line(0))
        tkinter.mainloop()

    def draw_line(self, y):
        _, pixeldata = self.mandel.calc_photoimage_line(y, res_x, res_y)
        self.img.put(pixeldata, (0, y))
        if y < res_y:
            self.root.after_idle(lambda: self.draw_line(y+1))
        else:
            duration = time.time() - self.start_time
            print("Calculation took: %.2f seconds" % duration)


if __name__ == "__main__":
    window = MandelWindow()



"""
This is a subscriber meant for the 'weather' messages example.
It uses a callback to process incoming messages.
"""
from __future__ import print_function
import sys
import Pyro4
from messagebus.messagebus import Subscriber
from Pyro4.util import excepthook

sys.excepthook = excepthook
if sys.version_info < (3, 0):
    input = raw_input

Pyro4.config.AUTOPROXY = True


@Pyro4.expose
class Subber(Subscriber):
    def consume_message(self, topic, message):
        # This callback-method is called automatically when a message arrives on the bus.
        print("\nGOT MESSAGE:")
        print("   topic:", topic)
        print("   msgid:", message.msgid)
        print(" created:", message.created)
        print("    data:", message.data)


hostname = input("hostname to bind on (empty=localhost): ").strip() or "localhost"

# create a messagebus subscriber that uses automatic message retrieval (via a callback)
subber = Subber()
d = Pyro4.Daemon(host=hostname)
d.register(subber)

topics = subber.bus.topics()
print("Topics on the bus: ", topics)
print("Subscribing to weather-forecast.")

subber.bus.subscribe("weather-forecast", subber)
# note: we subscribe on the bus *after* registering the subber as a Pyro object
# this results in Pyro automatically making a proxy for the subber
print("Subscribed on weather-forecast")
d.requestLoop()



from __future__ import print_function
import sys

import Pyro4


if sys.version_info < (3, 0):
    input = raw_input


@Pyro4.expose
class Viewer(object):
    def quote(self, market, symbol, value):
        print("{0}.{1}: {2}".format(market, symbol, value))


def main():
    viewer = Viewer()
    with Pyro4.Daemon() as daemon:
        daemon.register(viewer)
        aggregator = Pyro4.Proxy("PYRONAME:example.stockquote-old.aggregator")
        print("Available stock symbols:", aggregator.available_symbols())
        symbols = input("Enter symbols you want to view (comma separated):")
        symbols = [symbol.strip() for symbol in symbols.split(",")]
        aggregator.view(viewer, symbols)
        print("Viewer listening on symbols", symbols)
        daemon.requestLoop()


if __name__ == "__main__":
    main()



from __future__ import print_function


class Aggregator(object):
    def __init__(self):
        self.viewers = {}
        self.symbols = []

    def add_symbols(self, symbols):
        self.symbols.extend(symbols)

    def available_symbols(self):
        return self.symbols

    def view(self, viewer, symbols):
        print("aggregator gets a new viewer, for symbols:", symbols)
        self.viewers[viewer] = symbols

    def quotes(self, market, stockquotes):
        for symbol, value in stockquotes.items():
            for viewer, symbols in self.viewers.items():
                if symbol in symbols:
                    viewer.quote(market, symbol, value)


def main(stockmarkets):
    aggregator = Aggregator()
    for market in stockmarkets:
        aggregator.add_symbols(market.symbols())
        market.listener(aggregator)
    if not aggregator.available_symbols():
        raise ValueError("no symbols found!")
    print("aggregated symbols:", aggregator.available_symbols())
    return aggregator



from __future__ import print_function
import time
import Pyro4


with Pyro4.core.Proxy("PYRONAME:example.oneway") as serv:
    print("starting server using a oneway call")
    serv.oneway_start(6)
    print("doing some more oneway calls inbetween")
    serv.nothing()
    serv.nothing()
    serv.nothing()
    serv.nothing()

    time.sleep(2)
    print("\nNow contacting the server to see if it's done.")
    print("we are faster, so you should see a few attempts,")
    print("until the server is finished.")
    while True:
        print("server done?")
        if serv.ready():
            print("yes!")
            break
        else:
            print("no, trying again")
            time.sleep(1)

    print("getting the result from the server: %s" % serv.result())
    print("\nCalling oneway work method, server will continue working while we are done (check the server console output).")
    serv.oneway_work()



from __future__ import print_function
import Pyro4.core


class Safe(object):
    @Pyro4.expose
    def echo(self, message):
        print("got message:", message)
        return "hi!"


Pyro4.config.SSL = True
Pyro4.config.SSL_REQUIRECLIENTCERT = True   # enable 2-way ssl
Pyro4.config.SSL_SERVERCERT = "../../certs/server_cert.pem"
Pyro4.config.SSL_SERVERKEY = "../../certs/server_key.pem"
Pyro4.config.SSL_CACERTS = "../../certs/client_cert.pem"    # to make ssl accept the self-signed client cert
print("SSL enabled (2-way).")


class CertValidatingDaemon(Pyro4.core.Daemon):
    def validateHandshake(self, conn, data):
        cert = conn.getpeercert()
        if not cert:
            raise Pyro4.errors.CommunicationError("client cert missing")
        # note: hostname and expiry date validation is already successfully performed by the SSL layer itself
        # not_before = datetime.datetime.utcfromtimestamp(ssl.cert_time_to_seconds(cert["notBefore"]))
        # print("not before:", not_before)
        # not_after = datetime.datetime.utcfromtimestamp(ssl.cert_time_to_seconds(cert["notAfter"]))
        # print("not after:", not_after)
        # today = datetime.datetime.now()
        # if today > not_after or today < not_before:
        #     raise Pyro4.errors.CommunicationError("cert not yet valid or expired")
        if cert["serialNumber"] != "9BFD9872D96F066C":
            raise Pyro4.errors.CommunicationError("cert serial number incorrect")
        issuer = dict(p[0] for p in cert["issuer"])
        subject = dict(p[0] for p in cert["subject"])
        if issuer["organizationName"] != "Razorvine.net":
            # issuer is not often relevant I guess, but just to show that you have the data
            raise Pyro4.errors.CommunicationError("cert not issued by Razorvine.net")
        if subject["countryName"] != "NL":
            raise Pyro4.errors.CommunicationError("cert not for country NL")
        if subject["organizationName"] != "Razorvine.net":
            raise Pyro4.errors.CommunicationError("cert not for Razorvine.net")
        print("(SSL client cert is ok: serial={ser}, subject={subj})"
              .format(ser=cert["serialNumber"], subj=subject["organizationName"]))
        return super(CertValidatingDaemon, self).validateHandshake(conn, data)


d = CertValidatingDaemon()
uri = d.register(Safe)
print("server uri:", uri)
d.requestLoop()



from __future__ import print_function
import time

import Pyro4


# set the oneway behavior to run inside a new thread, otherwise the client stalls.
# this is the default, but I've added it here just for clarification.
Pyro4.config.ONEWAY_THREADED = True


@Pyro4.expose
class Server(object):
    def __init__(self):
        self.busy = False

    @Pyro4.oneway
    def oneway_start(self, duration):
        print("start request received. Starting work...")
        self.busy = True
        for i in range(duration):
            time.sleep(1)
            print(duration - i)
        print("work is done!")
        self.busy = False

    def ready(self):
        print("ready status requested (%r)" % (not self.busy))
        return not self.busy

    def result(self):
        return "The result :)"

    def nothing(self):
        print("nothing got called, doing nothing")

    @Pyro4.oneway
    def oneway_work(self):
        for i in range(10):
            print("work work..", i+1)
            time.sleep(1)
        print("work's done!")


# main program

Pyro4.Daemon.serveSimple({
    Server: "example.oneway"
})



from __future__ import print_function

from awesome_thirdparty_library import AwesomeClass
import Pyro4


# create adapter class that only exposes what should be accessible,
# and calls into the library class from there:

class AwesomeAdapterClass(AwesomeClass):
    @Pyro4.expose
    def method(self, arg):
        print("Adapter class is called...")
        return super(AwesomeAdapterClass, self).method(arg)

    @Pyro4.expose
    def weird(self):
        result = super(AwesomeAdapterClass, self).weird()
        # we have full control over what is returned and can turn the custom
        # result class into a normal string value that has no issues traveling over the wire
        return "weird " + result.value


with Pyro4.Daemon() as daemon:
    # register the adapter class instead of the library class itself:
    uri = daemon.register(AwesomeAdapterClass, "example.thirdpartylib")
    print("adapter class registered, uri: ", uri)
    daemon.requestLoop()



from __future__ import print_function
from time import sleep
import sys

import Pyro4
import Pyro4.errors


# client side will not have timeout
Pyro4.config.COMMTIMEOUT = 0

# Not using auto-retry feature
Pyro4.config.MAX_RETRIES = 0

obj = Pyro4.core.Proxy("PYRONAME:example.autoretry")
print("Calling remote function 1st time (create connection)")
obj.add(1, 1)
print("Calling remote function 2nd time (not timed out yet)")
obj.add(2, 2)
print("Sleeping 1 second...")
sleep(1)
print("Calling remote function 3rd time (will raise an exception)")
try:
    obj.add(3, 3)
except Exception as e:
    print("Got exception %r as expected." % repr(e))

print("Now, let's enable the auto retry")
obj._pyroRelease()
obj._pyroMaxRetries = 2

print("Calling remote function 1st time (create connection)")
obj.add(1, 1)
print("Calling remote function 2nd time (not timed out yet)")
obj.add(2, 2)
print("Sleeping 1 second...")
sleep(1)
print("Calling remote function 3rd time (will not raise any exceptions)")
obj.add(3, 3)



# ascii animation of zooming a mandelbrot fractal,  z=z^2+c
from __future__ import print_function, division
import os
import time
import platform
from server import Mandelbrot

res_x = 100
res_y = 40


def screen(start, width):
    mandel = Mandelbrot()
    dr = width / res_x
    di = dr*(res_x/res_y)
    di *= 0.8   # aspect ratio correction
    lines = mandel.calc_lines(start, res_x, dr, di, 0, res_y)
    return "\n".join(x[1] for x in lines)


def cls():
    if platform.platform().startswith("Windows"):
        os.system("cls")
    else:
        print(chr(27)+"[2J"+chr(27)+"[1;1H", end="")  # ansi clear screen
    

def zoom():
    start = -2.0-1.0j
    width = 3.0
    duration = 30.0
    wallclock_start = time.time()
    frames = 0
    cls()
    print("This is a mandelbrot zoom animation running without Pyro, in a single Python process.")
    time.sleep(2)
    while True:
        time_passed = time.time() - wallclock_start
        if time_passed >= duration:
            break
        actual_width = width * (1-time_passed/duration/1.1)
        actual_start = start + (0.06-0.002j)*time_passed
        frame = screen(actual_start, actual_width)
        cls()
        fps = frames/time_passed if time_passed > 0 else 0
        print("%.1f FPS time=%.2f width=%.2f" % (fps, time_passed, actual_width))
        print(frame)
        frames += 1
    print("Final FPS: %.2f" % fps)


if __name__ == "__main__":
    zoom()



from __future__ import print_function
import sys

import Pyro4


if sys.version_info < (3, 0):
    input = raw_input

uri = input("enter factory server object uri: ").strip()
factory = Pyro4.Proxy(uri)

# create several things.
print("Creating things.")
thing1 = factory.createSomething(1)
thing2 = factory.createSomething(2)
thing3 = factory.createSomething(3)

# interact with them on the server.
print("Speaking stuff.")
thing1.speak("I am the first")
thing2.speak("I am second")
thing3.speak("I am last then...")



from __future__ import print_function
import time
import Pyro4.naming

print("Autoreconnect using Name Server.")


@Pyro4.expose
class TestClass(object):
    def method(self, arg):
        print("Method called with %s" % arg)
        print("You can now try to stop this server with ctrl-C/ctrl-Break")
        time.sleep(1)


# If we reconnect the object, it has to have the same objectId as before.
# for this example, we rely on the Name Server registration to get our old id back.
# If we KNOW 100% that PYRONAME-uris are the only thing used to access our
# object, we could skip all this and just register as usual.
# That works because the proxy, when reconnecting, will do a new nameserver lookup
# and receive the new object uri back. This REQUIRES:
#   - clients will never connect using a PYRO-uri
#   - client proxy._pyroBind() is never called
# BUT for sake of example, and because we really cannot guarantee the above,
# here we go for the safe route and reuse our previous object id.

ns = Pyro4.naming.locateNS()
try:
    existing = ns.lookup("example.autoreconnect")
    print("Object still exists in Name Server with id: %s" % existing.object)
    print("Previous daemon socket port: %d" % existing.port)
    # start the daemon on the previous port
    daemon = Pyro4.core.Daemon(port=existing.port)
    # register the object in the daemon with the old objectId
    daemon.register(TestClass, objectId=existing.object)
except Pyro4.errors.NamingError:
    print("There was no previous registration in the name server.")
    # just start a new daemon on a random port
    daemon = Pyro4.core.Daemon()
    # register the object in the daemon and let it get a new objectId
    # also need to register in name server because it's not there yet.
    uri = daemon.register(TestClass)
    ns.register("example.autoreconnect", uri)

print("Server started.")
daemon.requestLoop()

# note: we are not removing the name server registration when terminating!



from __future__ import print_function
import threading
import time
import random
import traceback


class DummyDatabase(object):
    """Key-value datastore"""
    def __init__(self):
        self.storage = {}
        self.allowed_users = ["user123", "admin"]

    def connect(self, user):
        return Connection(self, user)

    def __setitem__(self, key, value):
        time.sleep(random.random()/10)   # artificial delay
        self.storage[key] = value

    def __getitem__(self, item):
        time.sleep(random.random()/10)   # artificial delay
        return self.storage[item]


class Connection(object):
    """
    Connection to the key-value datastore with artificial limitation
    that only a single thread may use the connection at the same time
    """
    def __init__(self, db, user=None):
        self.db = db
        self.user = user
        self.lock = threading.RLock()

    def store(self, key, value, user=None):
        user = user or self.user
        assert user in self.db.allowed_users, "access denied"
        if self.lock.acquire(blocking=False):
            print("DB: user %s stores: %s = %s" % (user, key, value))
            self.db[key] = value
            self.lock.release()
        else:
            raise RuntimeError("ERROR: concurrent connection access (write) by multiple different threads")

    def retrieve(self, key, user=None):
        user = user or self.user
        assert user in self.db.allowed_users, "access denied"
        if self.lock.acquire(blocking=False):
            print("DB: user %s retrieve: %s" % (user, key))
            value = self.db[key]
            self.lock.release()
            return value
        else:
            raise RuntimeError("ERROR: concurrent connection access (read) by multiple different threads")


if __name__ == "__main__":
    # first single threaded access
    db = DummyDatabase()
    conn = db.connect("user123")
    for i in range(5):
        conn.store("amount", 100+i)
        conn.retrieve("amount")

    # now multiple threads, should crash

    class ClientThread(threading.Thread):
        def __init__(self, conn):
            super(ClientThread, self).__init__()
            self.conn = conn
            self.daemon = True
        def run(self):
            for i in range(5):
                try:
                    self.conn.store("amount", 100+i)
                except Exception:
                    traceback.print_exc()
                try:
                    self.conn.retrieve("amount")
                except Exception:
                    traceback.print_exc()

    client1 = ClientThread(conn)
    client2 = ClientThread(conn)
    client1.start()
    client2.start()
    time.sleep(0.1)
    client1.join()
    client2.join()



"""
Tests for the built-in test echo server.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import time
import unittest
from threading import Thread, Event
import Pyro4.test.echoserver as echoserver
import Pyro4.util
from Pyro4.configuration import config


class EchoServerThread(Thread):
    def __init__(self):
        super(EchoServerThread, self).__init__()
        self.setDaemon(True)
        self.started = Event()
        self.echodaemon = self.echoserver = self.uri = None

    def run(self):
        self.echodaemon, self.echoserver, self.uri = echoserver.main(args=["-q"], returnWithoutLooping=True)
        self.started.set()
        self.echodaemon.requestLoop(loopCondition=lambda: not self.echoserver._must_shutdown)


class TestEchoserver(unittest.TestCase):
    def setUp(self):
        self.echoserverthread = EchoServerThread()
        self.echoserverthread.start()
        self.echoserverthread.started.wait()
        self.uri = self.echoserverthread.uri

    def tearDown(self):
        self.echoserverthread.echodaemon.shutdown()
        time.sleep(0.02)
        self.echoserverthread.join()
        config.SERVERTYPE = "thread"

    def testExposed(self):
        e = Pyro4.test.echoserver.EchoServer()
        self.assertTrue(hasattr(e, "_pyroExposed"))

    def testEcho(self):
        with Pyro4.core.Proxy(self.uri) as echo:
            try:
                self.assertEqual("hello", echo.echo("hello"))
                self.assertEqual(None, echo.echo(None))
                self.assertEqual([1, 2, 3], echo.echo([1, 2, 3]))
            finally:
                echo.shutdown()

    def testError(self):
        with Pyro4.core.Proxy(self.uri) as echo:
            try:
                echo.error()
                self.fail("expected exception")
            except Exception as x:
                tb = "".join(Pyro4.util.getPyroTraceback())
                self.assertIn("Remote traceback", tb)
                self.assertIn("ValueError", tb)
                self.assertEqual("expected error from echoserver error() method", str(x))
            try:
                echo.error_with_text()
                self.fail("expected exception")
            except Exception as x:
                tb = "".join(Pyro4.util.getPyroTraceback())
                self.assertIn("Remote traceback", tb)
                self.assertIn("ValueError", tb)
                self.assertEqual("the message of the error", str(x))

    def testGenerator(self):
        with Pyro4.core.Proxy(self.uri) as echo:
            remotegenerator = echo.generator()
            self.assertIsInstance(remotegenerator, Pyro4.core._StreamResultIterator)
            next(remotegenerator)
            next(remotegenerator)
            next(remotegenerator)
            with self.assertRaises(StopIteration):
                next(remotegenerator)


if __name__ == "__main__":
    unittest.main()



# This is an AWESOME LIBRARY.
# You can use its AWESOME CLASSES to do Great Things.
# The author however DOESN'T allow you to CHANGE the source code and taint it with Pyro decorators!


class WeirdReturnType(object):
    def __init__(self, value):
        self.value = value


class AwesomeClass(object):
    def method(self, arg):
        print("Awesome object is called with: ", arg)
        return "awesome"

    def private(self):
        print("This should be a private method...")
        return "boo"

    def weird(self):
        print("Weird!")
        return WeirdReturnType("awesome")



"""
Support code for the test suite.
There's some Python 2.x <-> 3.x compatibility code here.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import sys
import pickle
import threading
from Pyro4 import errors, core, expose, behavior, current_context
from Pyro4.configuration import config


__all__ = ["tobytes", "tostring", "unicode", "unichr", "basestring", "StringIO",
           "NonserializableError", "MyThingPartlyExposed", "MyThingFullExposed",
           "MyThingExposedSub", "MyThingPartlyExposedSub", "ConnectionMock",
           "AtomicCounter", "ResourceService", "Resource"]


config.reset(False)   # reset the config to default

if sys.version_info < (3, 0):
    # noinspection PyUnresolvedReferences
    from StringIO import StringIO

    def tobytes(string, encoding=None):
        return string

    def tostring(bytes):
        return bytes

    unicode = unicode
    unichr = unichr
    basestring = basestring
else:
    from io import StringIO

    def tobytes(string, encoding="iso-8859-1"):
        return bytes(string, encoding)

    def tostring(bytes, encoding="utf-8"):
        return str(bytes, encoding)

    unicode = str
    unichr = chr
    basestring = str


class NonserializableError(Exception):
    def __reduce__(self):
        raise pickle.PicklingError("to make this error non-serializable")


class MyThingPartlyExposed(object):
    c_attr = "hi"
    propvalue = 42
    _private_attr1 = "hi"
    __private_attr2 = "hi"
    name = ""

    def __init__(self, name="dummy"):
        self.name = name

    def __eq__(self, other):
        if type(other) is MyThingPartlyExposed:
            return self.name == other.name
        return False

    def method(self, arg, default=99, **kwargs):
        pass

    @staticmethod
    def staticmethod(arg):
        pass

    @classmethod
    def classmethod(cls, arg):
        pass

    def __dunder__(self):
        pass

    def __private(self):
        pass

    def _private(self):
        pass

    @core.expose
    @property
    def prop1(self):
        return self.propvalue

    @core.expose
    @prop1.setter
    def prop1(self, value):
        self.propvalue = value

    @core.expose
    @property
    def readonly_prop1(self):
        return self.propvalue

    @property
    def prop2(self):
        return self.propvalue

    @prop2.setter
    def prop2(self, value):
        self.propvalue = value

    @core.oneway
    @core.expose
    def oneway(self, arg):
        pass

    @core.expose
    def exposed(self):
        pass

    __hash__ = object.__hash__


@core.expose
class MyThingFullExposed(object):
    """this is the same as MyThingPartlyExposed but the whole class should be exposed"""
    c_attr = "hi"
    propvalue = 42
    _private_attr1 = "hi"
    __private_attr2 = "hi"
    name = ""

    def __init__(self, name="dummy"):
        self.name = name    # note: not affected by @expose, only real properties are

    def __eq__(self, other):
        if type(other) is MyThingFullExposed:
            return self.name == other.name
        return False

    def method(self, arg, default=99, **kwargs):
        pass

    @staticmethod
    def staticmethod(arg):
        pass

    @classmethod
    def classmethod(cls, arg):
        pass

    def __dunder__(self):
        pass

    def __private(self):
        pass

    def _private(self):
        pass

    @property
    def prop1(self):
        return self.propvalue

    @prop1.setter
    def prop1(self, value):
        self.propvalue = value

    @property
    def readonly_prop1(self):
        return self.propvalue

    @property
    def prop2(self):
        return self.propvalue

    @prop2.setter
    def prop2(self, value):
        self.propvalue = value

    @core.oneway
    def oneway(self, arg):
        pass

    def exposed(self):
        pass

    __hash__ = object.__hash__


@core.expose
class MyThingExposedSub(MyThingFullExposed):
    def sub_exposed(self):
        pass

    def sub_unexposed(self):
        pass

    @core.oneway
    def oneway2(self):
        pass


class MyThingPartlyExposedSub(MyThingPartlyExposed):
    @core.expose
    def sub_exposed(self):
        pass

    def sub_unexposed(self):
        pass

    @core.oneway
    def oneway2(self):
        pass


class ConnectionMock(object):
    def __init__(self, initial_msg=None):
        self.keep_open = False
        if not initial_msg:
            self.received = b""
        elif isinstance(initial_msg, (str, bytes)):
            self.received = initial_msg
        else:
            self.received = initial_msg.to_bytes()   # it's probably a Message object

    def send(self, data):
        self.received += data

    def recv(self, datasize):
        chunk = self.received[:datasize]
        self.received = self.received[datasize:]
        if len(chunk) < datasize:
            raise errors.ConnectionClosedError("receiving: not enough data")
        return chunk


class AtomicCounter(object):
    def __init__(self, value=0):
        self.__initial = value
        self.__value = value
        self.__lock = threading.Lock()

    def reset(self):
        self.__value = self.__initial

    def incr(self, amount=1):
        with self.__lock:
            self.__value += amount
            return self.__value

    def decr(self, amount=1):
        with self.__lock:
            self.__value -= amount
            return self.__value

    @property
    def value(self):
        return self.__value


class Resource(object):
    # a fictional resource that gets allocated and must be freed again later.
    def __init__(self, name, collection):
        self.name = name
        self.collection = collection
        self.close_called = False

    def close(self):
        # Pyro will call this on a tracked resource once the client's connection gets closed!
        self.collection.discard(self)
        self.close_called = True


@expose
@behavior(instance_mode="single")
class ResourceService(object):
    def __init__(self):
        self.resources = set()      # the allocated resources

    def allocate(self, name):
        resource = Resource(name, self.resources)
        self.resources.add(resource)
        current_context.track_resource(resource)

    def free(self, name):
        resources = {r for r in self.resources if r.name == name}
        self.resources -= resources
        for r in resources:
            r.close()
            current_context.untrack_resource(r)

    def list(self):
        return [r.name for r in self.resources]



from __future__ import print_function
import sys
import time

from Pyro4.util import getPyroTraceback
import Pyro4


if sys.version_info < (3, 0):
    input = raw_input

NUMBER_OF_LOOPS = 20000

uri = input("enter server object uri: ").strip()
p = Pyro4.Proxy(uri)

# First, we do a loop of N normal remote calls on the proxy
# We time the loop and validate the computation result
print("Normal remote calls...")
begin = time.time()
total = 0
p.printmessage("beginning normal calls")
for i in range(NUMBER_OF_LOOPS):
    total += p.multiply(7, 6)
    total += p.add(10, 20)
p.printmessage("end of normal calls")
assert total == (NUMBER_OF_LOOPS * (7 * 6 + 10 + 20))  # check
duration = time.time() - begin
print("that took {0:.2f} seconds ({1:.0f} calls/sec)".format(duration, NUMBER_OF_LOOPS * 2.0 / duration))
duration_normal = duration


# Now we do the same loop of N remote calls but this time we use
# the batched calls proxy. It collects all calls and processes them
# in a single batch. For many subsequent calls on the same proxy this
# is much faster than doing all calls individually.
# (but it has a few limitations and requires changes to your code)
print("\nBatched remote calls...")
begin = time.time()
batch = Pyro4.batch(p)  # get a batched call proxy for 'p'
batch.printmessage("beginning batch #1")
for i in range(NUMBER_OF_LOOPS):
    batch.multiply(7, 6)  # queue a call, note that it returns 'None' immediately
    batch.add(10, 20)  # queue a call, note that it returns 'None' immediately
batch.printmessage("end of batch #1")
print("processing the results...")
total = 0
result = batch()  # execute the batch of remote calls, it returns a generator that produces all results in sequence
for r in result:
    total += r
duration = time.time() - begin
assert total == (NUMBER_OF_LOOPS * (7 * 6 + 10 + 20))  # check
print("total time taken {0:.2f} seconds ({1:.0f} calls/sec)".format(duration, NUMBER_OF_LOOPS * 2.0 / duration // 100 * 100))
print("batched calls were {0:.1f} times faster than normal remote calls".format(duration_normal / duration))

# Now we do another loop of batched calls, but this time oneway (no results).
print("\nOneway batched remote calls...")
begin = time.time()
batch = Pyro4.batch(p)  # get a batched call proxy for 'p'
batch.printmessage("beginning batch #2")
for i in range(NUMBER_OF_LOOPS):
    batch.multiply(7, 6)  # queue a call, note that it returns 'None' immediately
    batch.add(10, 20)  # queue a call, note that it returns 'None' immediately
batch.delay(2)  # queue a delay of 2 seconds (but we won't notice)
batch.printmessage("end of batch #2")
print("executing batch, there will be no result values. Check server to see printed messages...")
result = batch(oneway=True)  # execute the batch of remote calls, oneway, will return None
assert result is None
duration = time.time() - begin
print("total time taken {0:.2f} seconds ({1:.0f} calls/sec)".format(duration, NUMBER_OF_LOOPS * 2.0 / duration // 100 * 100))
print("oneway batched calls were {0:.1f} times faster than normal remote calls".format(duration_normal / duration))

# Batches can be executed asynchronous as well
print("\nBatched remote calls, asynchronous...")
batch = Pyro4.batch(p)  # get a batched call proxy for 'p'
batch.printmessage("beginning batch #3")
batch.multiply(7, 6)  # queue a call, note that it returns 'None' immediately
batch.add(10, 20)  # queue a call, note that it returns 'None' immediately
batch.delay(2)  # queue a delay, but this doesn't matter with asynchronous proxy
batch.printmessage("end of batch #3")
print("executing the batch... (should return immediately because async)")
asyncresult = batch(asynchronous=True)  # execute the batch, asynchronously (return immediately)
print("processing the results...(should wait until async results become available)")
results = list(asyncresult.value)
print("results=", results)


# Show what happens when one of the methods in a batch generates an error.
# (the batch is aborted and the error is raised locally again).
# Btw, you can re-use a batch proxy once you've called it and processed the results.
print("\nBatch with an error. Dividing a number by decreasing divisors...")
for d in range(3, -3, -1):  # divide by 3,2,1,0,-1,-2,-3... but 0 will be a problem ;-)
    batch.divide(100, d)
print("getting results...")
divisor = 3
try:
    for result in batch():
        print("100//%d = %d" % (divisor, result))
        divisor -= 1
        # this will raise the proper zerodivision exception once we're about
        # to process the batch result from the divide by 0 call.
except ZeroDivisionError:
    print("A divide by zero error occurred during the batch! (expected)")
    print("".join(getPyroTraceback()))



from __future__ import print_function
import sys
import re
import unittest

try:
    # try setuptools first, to get access to build_sphinx and test commands
    from setuptools import setup

    using_setuptools = True
except ImportError:
    from distutils.core import setup

    using_setuptools = False


def pyro_test_suite():
    testloader = unittest.TestLoader()
    testsuite = testloader.discover("tests/PyroTests", pattern="test*.py")
    return testsuite


if __name__ == '__main__':
    with open("src/Pyro4/constants.py") as constants_file:
        # extract the VERSION definition from the Pyro4.constants module without importing it
        version_line = next(line for line in constants_file if line.startswith("VERSION"))
        pyro4_version = re.match("VERSION ?= ?['\"](.+)['\"]", version_line).group(1)
    print('Pyro version = %s' % pyro4_version)

    setup(
        name="Pyro4",
        version=pyro4_version,
        license="MIT",
        description="distributed object middleware for Python (RPC)",
        long_description="""Pyro means PYthon Remote Objects.
It is a library that enables you to build applications in which
objects can talk to eachother over the network, with minimal programming effort.
You can just use normal Python method calls, with almost every possible parameter
and return value type, and Pyro takes care of locating the right object on the right
computer to execute the method. It is designed to be very easy to use, and to
generally stay out of your way. But it also provides a set of powerful features that
enables you to build distributed applications rapidly and effortlessly.
Pyro is a pure Python library and runs on many different platforms and Python versions.

The source code repository is on Github: https://github.com/irmen/Pyro4

The documentation can be found here: http://pyro4.readthedocs.io
""",
        author="Irmen de Jong",
        author_email="irmen@razorvine.net",
        keywords=["distributed objects", "RPC", "remote method call", "IPC"],
        url="http://pyro4.readthedocs.io",
        package_dir={'': 'src'},
        packages=['Pyro4', 'Pyro4.socketserver', 'Pyro4.test', 'Pyro4.utils'],
        scripts=[],
        platforms="any",
        test_suite="setup.pyro_test_suite",
        install_requires=[
            "serpent>=1.27,<1.30 ; python_version<'3.2'",
            "serpent>=1.27 ; python_version>='3.2'",
        ],
        extras_require={
            ":python_version<'3.4'": ["selectors34"]
        },
        requires=["serpent"],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Development Status :: 6 - Mature",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Natural Language :: Dutch",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Topic :: Software Development :: Object Brokering",
            "Topic :: System :: Distributed Computing",
            "Topic :: System :: Networking"
        ],
        entry_points={
            'console_scripts': [
                'pyro4-ns=Pyro4.naming:main',
                'pyro4-nsc=Pyro4.nsc:main',
                'pyro4-test-echoserver=Pyro4.test.echoserver:main',
                'pyro4-check-config=Pyro4.configuration:main',
                'pyro4-flameserver=Pyro4.utils.flameserver:main',
                'pyro4-httpgateway=Pyro4.utils.httpgateway:main'
            ]
        },
        options={"install": {"optimize": 0}}
    )

    if len(sys.argv) >= 2 and sys.argv[1].startswith("install"):
        print("\nOnly the Pyro library has been installed (version %s)." % pyro4_version)
        print("If you want to install the tests, the examples, and/or the manual,")
        print("you have to copy them manually to the desired location.")



from __future__ import print_function
import Pyro4
import threading


@Pyro4.expose
class EchoServer(object):
    def echo(self, message):
        ctx = Pyro4.current_context
        print("\nGot Message:", message)
        print("  thread: ", threading.current_thread().ident)
        print("  obj.pyroid: ", self._pyroId)
        print("  obj.daemon: ", self._pyroDaemon)
        print("  context.client: ", ctx.client)
        print("  context.client_sock_addr: ", ctx.client_sock_addr)
        print("  context.seq: ", ctx.seq)
        print("  context.msg_flags: ", ctx.msg_flags)
        print("  context.serializer_id: ", ctx.serializer_id)
        print("  context.correlation_id:", ctx.correlation_id)
        if "XYZZ" in ctx.annotations:
            print("  custom annotation 'XYZZ':", ctx.annotations["XYZZ"])
        return message

    @Pyro4.oneway
    def oneway(self, message):
        return self.echo(message)


class CustomDaemon(Pyro4.Daemon):
    def annotations(self):
        return {"DDAA": b"custom response annotation set by the daemon"}


with CustomDaemon() as daemon:
    daemon._pyroHmacKey = b"secr3t_k3y"
    uri = daemon.register(EchoServer, "example.context")  # provide a logical name ourselves
    print("Server is ready. You can use the following URI to connect:")
    print(uri)
    daemon.requestLoop()



from __future__ import print_function
import threading
import time
import sys
import Pyro4


if sys.version_info < (3, 0):
    input = raw_input


def get_user_token():
    return "user123"


class CustomAnnotationProxy(Pyro4.Proxy):
    # override the method that adds annotations and add our own custom user token annotation
    def _pyroAnnotations(self):
        return {"USER": get_user_token().encode("utf-8")}


class DbAccessor(threading.Thread):
    def __init__(self, uri):
        super(DbAccessor, self).__init__()
        self.proxy = CustomAnnotationProxy(uri)
        self.daemon = True

    def run(self):
        for i in range(3):
            try:
                self.proxy.store("number", 100+i)
                num = self.proxy.retrieve("number")
                print("[%s] num=%s" % (self.name, num))
            except Exception:
                import traceback
                traceback.print_exc()


print("\n***** Sequential access using multiple proxies on the Session-Bound Database... (no issues)")
with CustomAnnotationProxy("PYRONAME:example.usersession.sessiondb") as p1,\
        CustomAnnotationProxy("PYRONAME:example.usersession.sessiondb") as p2:
    p1.store("number", 42)
    p1.retrieve("number")
    p2.store("number", 43)
    p2.retrieve("number")

print("\n***** Sequential access using multiple proxies on the Singleton Database... (no issues)")
with CustomAnnotationProxy("PYRONAME:example.usersession.singletondb") as p1,\
        CustomAnnotationProxy("PYRONAME:example.usersession.singletondb") as p2:
    p1.store("number", 42)
    p1.retrieve("number")
    p2.store("number", 43)
    p2.retrieve("number")

print("\n***** Multiple concurrent proxies on the Session-Bound Database... (no issues)")
input("enter to start: ")
t1 = DbAccessor("PYRONAME:example.usersession.sessiondb")
t2 = DbAccessor("PYRONAME:example.usersession.sessiondb")
t1.start()
t2.start()
time.sleep(1)
t1.join()
t2.join()

print("\n***** Multiple concurrent proxies on the Singleton Database... (threading problem)")
input("enter to start: ")
t1 = DbAccessor("PYRONAME:example.usersession.singletondb")
t2 = DbAccessor("PYRONAME:example.usersession.singletondb")
t1.start()
t2.start()
time.sleep(1)
t1.join()
t2.join()



from Pyro4 import expose


class Resource(object):
    pass


@expose
class LaserPrinter(Resource):
    pass


@expose
class MatrixPrinter(Resource):
    pass


@expose
class PhotoPrinter(Resource):
    pass


@expose
class TapeStorage(Resource):
    pass


@expose
class DiskStorage(Resource):
    pass


@expose
class Telephone(Resource):
    pass


@expose
class Faxmachine(Resource):
    pass



"""
Tests for some Ironpython peculiarities.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import sys
import pickle
import unittest
import Pyro4.util


if sys.platform == "cli":

    class IronPythonWeirdnessTests(unittest.TestCase):
        def testExceptionWithAttrsPickle(self):
            # ironpython doesn't pickle exception attributes
            # Bug report is at https://github.com/IronLanguages/main/issues/943
            # Bug is still present in Ironpython 2.7.7
            ex = ValueError("some exception")
            ex.custom_attribute = 42
            ex2 = pickle.loads(pickle.dumps(ex))
            self.assertTrue(hasattr(ex, "custom_attribute"))
            self.assertFalse(hasattr(ex2, "custom_attribute"))  # custom attribute will be gone after pickling
            self.assertNotEqual(ex2, ex)  # the object won't be equal

        def testExceptionReduce(self):
            # ironpython doesn't pickle exception attributes
            # Bug report is at https://github.com/IronLanguages/main/issues/943
            # Bug is still present in Ironpython 2.7.7
            ex = ValueError("some exception")
            ex.custom_attribute = 42
            r = ex.__reduce__()
            # the reduce result should be:
            # (ValueError, ("some exception",), {"custom_attribute": 42})
            # but in Ironpython the custom attributes are not returned.
            self.assertNotEqual((ValueError, ("some exception",), {"custom_attribute": 42}), r)
            self.assertEqual((ValueError, ("some exception",)), r)

        def testTbFrame(self):
            # there's some stuff missing on traceback frames
            # this prevents a detailed stack trace to be printed by
            # the functions in util.py, for instance.
            def crash():
                a = 1
                b = 0
                return a // b

            try:
                crash()
            except:
                ex_t, ex_v, ex_tb = sys.exc_info()
                while ex_tb.tb_next:
                    ex_tb = ex_tb.tb_next
                self.assertIsNone(ex_tb.tb_frame.f_back)  # should not be none... :(

        def testExceptionArgs(self):
            x = ZeroDivisionError("division by zero", "arg1", "arg2")
            x.customattribute = 42
            Pyro4.util.fixIronPythonExceptionForPickle(x, True)
            arg = x.args[-1]
            self.assertIsInstance(arg, dict)
            self.assertTrue(arg["__ironpythonargs__"])
            self.assertEqual(42, arg["customattribute"])
            x = ZeroDivisionError("division by zero", "arg1", "arg2")
            x.args += ({"__ironpythonargs__": True, "customattribute2": 99},)
            Pyro4.util.fixIronPythonExceptionForPickle(x, False)
            self.assertEqual(99, x.customattribute2)


if __name__ == "__main__":
    unittest.main()



from __future__ import print_function
import sys

import Pyro4
from Pyro4.util import SerializerBase
import mycustomclasses


# use serpent
Pyro4.config.SERIALIZER = "serpent"


# register the special serialization hooks

def thingy_class_to_dict(obj):
    print("{serializer hook, converting to dict: %s}" % obj)
    return {
        "__class__": "waheeee-custom-thingy",
        "number-attribute": obj.number
    }


def thingy_dict_to_class(classname, d):
    print("{deserializer hook, converting to class: %s}" % d)
    return mycustomclasses.Thingy(d["number-attribute"])


def otherthingy_dict_to_class(classname, d):
    print("{deserializer hook, converting to class: %s}" % d)
    return mycustomclasses.OtherThingy(d["number"])

# for 'Thingy' we register both serialization and deserialization hooks
SerializerBase.register_class_to_dict(mycustomclasses.Thingy, thingy_class_to_dict)
SerializerBase.register_dict_to_class("waheeee-custom-thingy", thingy_dict_to_class)

# for 'OtherThingy' we only register a deserialization hook (and for serialization depend on serpent's default behavior)
SerializerBase.register_dict_to_class("mycustomclasses.OtherThingy", otherthingy_dict_to_class)


# regular pyro stuff
if sys.version_info < (3, 0):
    input = raw_input

uri = input("Enter the URI of the server object: ")
serv = Pyro4.core.Proxy(uri)
print("\nTransferring thingy...")
o = mycustomclasses.Thingy(42)
response = serv.method(o)
print("type of response object:", type(response))
print("response:", response)
print("\nTransferring otherthingy...")
o = mycustomclasses.OtherThingy(42)
response = serv.othermethod(o)
print("type of response object:", type(response))
print("response:", response)



# The server that doesn't use the Name Server.

from __future__ import print_function
import os

import Pyro4


class QuoteGen(object):
    @Pyro4.expose
    def quote(self):
        try:
            quote = os.popen('fortune').read()
            if len(quote) > 0:
                return quote
            return "This system cannot provide you a good fortune, install 'fortune'"
        except:
            return "This system knows no witty quotes :-("


with Pyro4.core.Daemon() as daemon:
    quote1 = QuoteGen()
    quote2 = QuoteGen()

    uri1 = daemon.register(quote1)  # let Pyro create a unique name for this one
    uri2 = daemon.register(quote2, "example.quotegen")  # provide a logical name ourselves

    print("QuoteGen is ready, not using the Name Server.")
    print("You can use the following two URIs to connect to me:")
    print(uri1)
    print(uri2)

    daemon.requestLoop()



from __future__ import print_function
from collections import defaultdict
import Pyro4

# note: the dispatcher doesn't know anything about the CustomData class from the customdata module!


@Pyro4.behavior(instance_mode="single")
class Dispatcher(object):
    def __init__(self):
        self.listeners = defaultdict(list)

    @Pyro4.expose
    def register(self, topic, listener):
        self.listeners[topic].append(listener)
        print("New listener for topic {} registered: {}".format(topic, listener._pyroUri))

    @Pyro4.expose
    def process_blob(self, blob):
        print("Dispatching blob with name:", blob.info)
        listeners = self.listeners.get(blob.info, [])
        for listener in listeners:
            listener.process_blob(blob)


Pyro4.Daemon.serveSimple({
    Dispatcher: "example.blobdispatcher"
})



from __future__ import print_function
import logging

import Pyro4


logging.basicConfig(level=logging.DEBUG)
logging.getLogger("Pyro4").setLevel(logging.DEBUG)

Pyro4.config.COMMTIMEOUT = 5.0
Pyro4.config.POLLTIMEOUT = 5.0  # only used for multiplexing server


class TestDisconnect(object):
    @Pyro4.expose
    def echo(self, arg):
        print("echo: ", arg)
        return arg


Pyro4.Daemon.serveSimple({
    TestDisconnect: "example.disconnect"
}, ns=False)



from __future__ import print_function
import logging
import sys
import Pyro4


# initialize the logger so you can see what is happening with the callback exception message:
logging.basicConfig(stream=sys.stderr, format="[%(asctime)s,%(name)s,%(levelname)s] %(message)s")
log = logging.getLogger("Pyro4")
log.setLevel(logging.WARNING)


class CallbackHandler(object):
    def crash(self):
        a = 1
        b = 0
        return a // b

    @Pyro4.expose
    def call1(self):
        print("\n\ncallback 1 received from server!")
        print("going to crash - you won't see the exception here, only on the server")
        return self.crash()

    @Pyro4.expose
    @Pyro4.callback
    def call2(self):
        print("\n\ncallback 2 received from server!")
        print("going to crash - but you will see the exception here too")
        return self.crash()


daemon = Pyro4.core.Daemon()
callback = CallbackHandler()
daemon.register(callback)

with Pyro4.core.Proxy("PYRONAME:example.callback2") as server:
    server.doCallback(callback)   # this is a oneway call, so we can continue right away

print("waiting for callbacks to arrive...")
print("(ctrl-c/break the program once it's done)")
daemon.requestLoop()



import random
import sys
import robot
import remote
import Pyro4


class DrunkenGameObserver(remote.GameObserver):
    @Pyro4.oneway
    @Pyro4.expose
    def world_update(self, iteration, world, robotdata):
        # change directions randomly
        if random.random() > 0.8:
            if random.random() >= 0.5:
                dx, dy = random.randint(-1, 1), 0
            else:
                dx, dy = 0, random.randint(-1, 1)
            if random.random() > 0.7:
                self.robot.emote("..Hic! *burp*")
            self.robot.change_direction((dx, dy))


class AngryGameObserver(remote.GameObserver):
    def __init__(self):
        super(AngryGameObserver, self).__init__()
        self.directions = [(1, 0), (0, 1), (-1, 0), (0, -1)]  # clockwise motion
        self.directioncounter = 0

    @Pyro4.oneway
    @Pyro4.expose
    def world_update(self, iteration, world, robotdata):
        # move in a loop yelling angry stuff
        if iteration % 50 == 0:
            self.robot.emote("I'll kill you all! GRR")
        if iteration % 10 == 0:
            self.directioncounter = (self.directioncounter + 1) % 4
            self.robot.change_direction(self.directions[self.directioncounter])


class ScaredGameObserver(remote.GameObserver):
    def __init__(self):
        super(ScaredGameObserver, self).__init__()
        # run to a corner
        self.direction = random.choice([(-1, -1), (1, -1), (1, 1), (-1, 1)])

    @Pyro4.oneway
    @Pyro4.expose
    def start(self):
        super(ScaredGameObserver, self).start()
        self.robot.change_direction(self.direction)

    @Pyro4.oneway
    @Pyro4.expose
    def world_update(self, iteration, world, robotdata):
        if iteration % 50 == 0:
            self.robot.emote("I'm scared!")


observers = {
    "drunk": DrunkenGameObserver,
    "angry": AngryGameObserver,
    "scared": ScaredGameObserver,
}


# register the Robot class with Pyro's serializers:
Pyro4.util.SerializerBase.register_class_to_dict(robot.Robot, robot.Robot.robot_to_dict)
Pyro4.util.SerializerBase.register_dict_to_class("robot.Robot", robot.Robot.dict_to_robot)


def main(args):
    if len(args) != 3:
        print("usage: client.py <robotname> <robottype>")
        print("   type is one of: %s" % list(observers.keys()))
        return
    name = args[1]
    observertype = args[2]
    with Pyro4.Daemon() as daemon:
        observer = observers[observertype]()
        daemon.register(observer)
        gameserver = Pyro4.Proxy("PYRONAME:example.robotserver")
        robot = gameserver.register(name, observer)
        robot.emote("Hi there! I'm here to kick your ass")
        observer.robot = robot
        print("Pyro server registered on %s" % daemon.locationStr)
        daemon.requestLoop()


if __name__ == "__main__":
    main(sys.argv)



#
#    Bank client.
#
#    The client searches the two banks and performs a set of operations.
#    (the banks are searched simply by listing a namespace prefix path)
#

from __future__ import print_function
import Pyro4.naming


# A bank client.
class client(object):
    def __init__(self, name):
        self.name = name

    def doBusiness(self, bank):
        print("\n*** %s is doing business with %s:" % (self.name, bank.name()))
        print("Creating account")
        try:
            bank.createAccount(self.name)
        except ValueError as x:
            print("Failed: %s" % x)
            print("Removing account and trying again")
            bank.deleteAccount(self.name)
            bank.createAccount(self.name)

        print("Deposit money")
        bank.deposit(self.name, 200.00)
        print("Deposit money")
        bank.deposit(self.name, 500.75)
        print("Balance=%.2f" % bank.balance(self.name))
        print("Withdraw money")
        bank.withdraw(self.name, 400.00)
        print("Withdraw money (overdraw)")
        try:
            bank.withdraw(self.name, 400.00)
        except ValueError as x:
            print("Failed: %s" % x)
        print("End balance=%.2f" % bank.balance(self.name))

        print("Withdraw money from non-existing account")
        try:
            bank.withdraw('GOD', 2222.22)
            print("!!! Succeeded?!? That is an error")
        except KeyError as x:
            print("Failed as expected: %s" % x)

        print("Deleting non-existing account")
        try:
            bank.deleteAccount('GOD')
            print("!!! Succeeded?!? That is an error")
        except KeyError as x:
            print("Failed as expected: %s" % x)


ns = Pyro4.naming.locateNS()

# list the available banks by looking in the NS for the given prefix path
banknames = [name for name in ns.list(prefix="example.banks.")]
if not banknames:
    raise RuntimeError('There are no banks to do business with!')

banks = []  # list of banks (proxies)
print()
for name in banknames:
    print("Contacting bank: %s" % name)
    uri = ns.lookup(name)
    banks.append(Pyro4.core.Proxy(uri))

# Different clients that do business with all banks
irmen = client('Irmen')
suzy = client('Suzy')

for bank in banks:
    irmen.doBusiness(bank)
    suzy.doBusiness(bank)

# List all accounts
print()
for bank in banks:
    print("The accounts in the %s:" % bank.name())
    accounts = bank.allAccounts()
    for name in accounts.keys():
        print("  %s : %.2f" % (name, accounts[name]))



from __future__ import print_function
import sys
import Pyro4

if sys.version_info < (3, 0):
    input = raw_input

Pyro4.config.SERIALIZER = "cloudpickle"   # can also use "dill"

uri = input("Uri of extended-pickle example server? ").strip()


class WorkerThing(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, servername):
        return "this is the result of object %s on server %s" % (self.name, servername)


def somework(servername):
    return "this is the result of function somework on server %s" % servername


with Pyro4.core.Proxy(uri) as p:
    print("sending a callable object to the server:")
    print("  ", p.work(WorkerThing("peter")))
    print("sending a function to the server:")
    print("  ", p.work(somework))
    print("sending a lambda to the server:")
    print("  ", p.work(lambda servername: "this is the result of the lambda function on server %s" % servername))



from __future__ import print_function
import sys
import time

import Pyro4


if sys.version_info < (3, 0):
    input = raw_input


def asyncFunction(value, increase=1):
    print(">>> async function called with value={0} increase={1}".format(value, increase))
    return value + increase


def asyncWithMoreArgs(a, b, extra=None):
    print(">>> async func called with some arguments: a={0}, b={1}, extra={2}".format(a, b, extra))
    return a + b


uri = input("enter async server object uri: ").strip()
proxy = Pyro4.Proxy(uri)

print("\n* async call with error:")
proxy._pyroAsync()
def resulthandler(result):
    print("RESULT: ", result)
def errorhandler(error):
    print("ERRORHANDLER: ", error)
asyncresult = proxy.divide(100, 0).iferror(errorhandler).then(resulthandler)
time.sleep(4)
print("^^^^ above an error message should be printed out...")
print("\n* async call with error, but without errorhandler:")
asyncresult = proxy.divide(100, 0).then(resulthandler)
time.sleep(4)
print("^^^^ above, no error message should be printed out...")
try:
    _ = asyncresult.value
    print("SHOULD RAISE ERROR INSTEAD!!")
except ZeroDivisionError as x:
    print("expected error occurred when trying to obtain the result:", x)

print("\n* async call with call chain:")
asyncresult = proxy.divide(100, 5)
# set a chain of callables to be invoked once the value is available
asyncresult.then(asyncFunction) \
    .then(asyncFunction) \
    .then(asyncFunction)
print("sleeping 4 seconds")
time.sleep(4)  # the call chain will be invoked during this sleep period
print("back from sleep")
# you can still access the final asyncresult.value. It should be 100/5+3=23
print("final value=", asyncresult.value)
assert asyncresult.value == 23

print("\n* async call with call chain that is set 'too late':")
asyncresult = proxy.divide(100, 5)
# set a chain of callables to be invoked once the value is available
# but we set it 'too late' (when the result is already available)
time.sleep(4)  # result will appear in 3 seconds
asyncresult.then(asyncFunction) \
    .then(asyncFunction) \
    .then(asyncFunction)
# let's see what the result value is, should be 100/5+3=23
print("final value=", asyncresult.value)
assert asyncresult.value == 23

print("\n* async call with call chain, where calls have extra arguments:")
asyncresult = proxy.multiply(5, 6)
# set a chain of callables to be invoked once the value is available
# the callable will be invoked like so:  function(asyncvalue, normalarg, kwarg=..., kwarg=...)
# (the value from the previous call is passed as the first argument to the next call)
asyncresult.then(asyncWithMoreArgs, 1) \
    .then(asyncWithMoreArgs, 2, extra=42) \
    .then(asyncWithMoreArgs, 3, extra="last one")
print("sleeping 1 second")
time.sleep(1)  # the call chain will be invoked during this sleep period
print("back from sleep")
# you can still access the final asyncresult.value. It should be 5*6+1+2+3=36
print("final value=", asyncresult.value)
assert asyncresult.value == 36

print("\n* async call with call chain, where calls are new non-async pyro calls:")
normalproxy = Pyro4.Proxy(uri)
asyncresult = proxy.divide(100, 5)
# set a chain of callables to be invoked once the value is available
# the callable will be invoked like so:  function(asyncvalue, kwarg=..., kwarg=...)
asyncresult.then(normalproxy.add, increase=1) \
    .then(normalproxy.add, increase=2) \
    .then(normalproxy.add, increase=3)
print("getting result value (will block until available)")
print("final value=", asyncresult.value)
assert asyncresult.value == 26  # 100/5+1+2+3=26

print("\ndone.")



"""
Tests for the daemon.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

from __future__ import print_function
import sys
import time
import socket
import uuid
import unittest
import Pyro4.core
import Pyro4.naming
import Pyro4.constants
import Pyro4.socketutil
import Pyro4.message
import Pyro4.util
from Pyro4.errors import DaemonError, PyroError, SerializeError
from Pyro4.configuration import config
from Pyro4.core import current_context
from testsupport import *


class MyObj(object):
    def __init__(self, arg):
        self.arg = arg

    def __eq__(self, other):
        return self.arg == other.arg

    __hash__ = object.__hash__


class CustomDaemonInterface(Pyro4.core.DaemonObject):
    def __init__(self, daemon):
        super(CustomDaemonInterface, self).__init__(daemon)

    def custom_daemon_method(self):
        return 42


class DaemonTests(unittest.TestCase):
    # We create a daemon, but notice that we are not actually running the requestloop.
    # 'on-line' tests are all taking place in another test, to keep this one simple.

    def setUp(self):
        config.POLLTIMEOUT = 0.1

    def sendHandshakeMessage(self, conn, correlation_id=None):
        ser = Pyro4.util.get_serializer_by_id(Pyro4.util.MarshalSerializer.serializer_id)
        data, _ = ser.serializeData({"handshake": "hello", "object": Pyro4.constants.DAEMON_NAME}, False)
        annotations = {"CORR": correlation_id.bytes} if correlation_id else None
        msg = Pyro4.message.Message(Pyro4.message.MSG_CONNECT, data, Pyro4.util.MarshalSerializer.serializer_id, 0, 99, annotations=annotations)
        msg.send(conn)

    def testSerializerConfig(self):
        self.assertIsInstance(config.SERIALIZERS_ACCEPTED, set)
        self.assertIsInstance(config.SERIALIZER, basestring)
        self.assertGreater(len(config.SERIALIZERS_ACCEPTED), 1)

    def testSerializerAccepted(self):
        self.assertIn("marshal", config.SERIALIZERS_ACCEPTED)
        self.assertNotIn("pickle", config.SERIALIZERS_ACCEPTED)
        self.assertNotIn("cloudpickle", config.SERIALIZERS_ACCEPTED)
        self.assertNotIn("dill", config.SERIALIZERS_ACCEPTED)
        with Pyro4.core.Daemon(port=0) as d:
            msg = Pyro4.message.Message(Pyro4.message.MSG_INVOKE, b"", Pyro4.util.MarshalSerializer.serializer_id, 0, 0, hmac_key=d._pyroHmacKey)
            cm = ConnectionMock(msg)
            d.handleRequest(cm)  # marshal serializer should be accepted
            msg = Pyro4.message.Message(Pyro4.message.MSG_INVOKE, b"", Pyro4.util.PickleSerializer.serializer_id, 0, 0, hmac_key=d._pyroHmacKey)
            cm = ConnectionMock(msg)
            try:
                d.handleRequest(cm)
                self.fail("should crash")
            except Pyro4.errors.ProtocolError as x:
                self.assertIn("serializer that is not accepted", str(x))
                pass
            msg = Pyro4.message.Message(Pyro4.message.MSG_INVOKE, b"", Pyro4.util.CloudpickleSerializer.serializer_id, 0, 0, hmac_key=d._pyroHmacKey)
            cm = ConnectionMock(msg)
            try:
                d.handleRequest(cm)
                self.fail("should crash")
            except Pyro4.errors.ProtocolError as x:
                self.assertTrue("no serializer available for id" in str(x) or "serializer that is not accepted" in str(x))
            msg = Pyro4.message.Message(Pyro4.message.MSG_INVOKE, b"", Pyro4.util.DillSerializer.serializer_id, 0, 0, hmac_key=d._pyroHmacKey)
            cm = ConnectionMock(msg)
            try:
                d.handleRequest(cm)
                self.fail("should crash")
            except Pyro4.errors.ProtocolError as x:
                self.assertTrue("no serializer available for id" in str(x) or "serializer that is not accepted" in str(x))

    def testDaemon(self):
        with Pyro4.core.Daemon(port=0) as d:
            hostname, port = d.locationStr.split(":")
            port = int(port)
            self.assertIn(Pyro4.constants.DAEMON_NAME, d.objectsById)
            self.assertEqual("PYRO:" + Pyro4.constants.DAEMON_NAME + "@" + d.locationStr, str(d.uriFor(Pyro4.constants.DAEMON_NAME)))
            # check the string representations
            expected = "<Pyro4.core.Daemon at 0x%x; %s - %s; 1 objects>" % (id(d), d.locationStr, Pyro4.socketutil.family_str(d.sock))
            self.assertEqual(expected, str(d))
            self.assertEqual(expected, unicode(d))
            self.assertEqual(expected, repr(d))
            sockname = d.sock.getsockname()
            self.assertEqual(port, sockname[1])
            daemonobj = d.objectsById[Pyro4.constants.DAEMON_NAME]
            daemonobj.ping()
            daemonobj.registered()
            try:
                daemonobj.shutdown()
                self.fail("should not succeed to call unexposed method")
            except AttributeError:
                pass

    def testDaemonCustomInterface(self):
        with Pyro4.core.Daemon(port=0, interface=CustomDaemonInterface) as d:
            obj = d.objectsById[Pyro4.constants.DAEMON_NAME]
            self.assertEqual(42, obj.custom_daemon_method())

    def testDaemonConnectedSocket(self):
        try:
            Pyro4.config.SERVERTYPE = "thread"
            with Pyro4.core.Daemon() as d:
                self.assertTrue("Thread" in d.transportServer.__class__.__name__)
            s1, s2 = socket.socketpair()
            with Pyro4.core.Daemon(connected_socket=s1) as d:
                self.assertTrue(d.locationStr=="./u:<<not-bound>>" or d.locationStr.startswith("127.0."))
                self.assertFalse("Thread" in d.transportServer.__class__.__name__)
                self.assertTrue("Existing" in d.transportServer.__class__.__name__)
            Pyro4.config.SERVERTYPE = "multiplex"
            with Pyro4.core.Daemon() as d:
                self.assertTrue("Multiplex" in d.transportServer.__class__.__name__)
            s1, s2 = socket.socketpair()
            with Pyro4.core.Daemon(connected_socket=s1) as d:
                self.assertTrue(d.locationStr=="./u:<<not-bound>>" or d.locationStr.startswith("127.0."))
                self.assertFalse("Multiplex" in d.transportServer.__class__.__name__)
                self.assertTrue("Existing" in d.transportServer.__class__.__name__)
        finally:
            Pyro4.config.SERVERTYPE = "thread"

    @unittest.skipUnless(hasattr(socket, "AF_UNIX"), "unix domain sockets required")
    def testDaemonUnixSocket(self):
        SOCKNAME = "test_unixsocket"
        with Pyro4.core.Daemon(unixsocket=SOCKNAME) as d:
            locationstr = "./u:" + SOCKNAME
            self.assertEqual(locationstr, d.locationStr)
            self.assertEqual("PYRO:" + Pyro4.constants.DAEMON_NAME + "@" + locationstr, str(d.uriFor(Pyro4.constants.DAEMON_NAME)))
            # check the string representations
            expected = "<Pyro4.core.Daemon at 0x%x; %s - Unix; 1 objects>" % (id(d), locationstr)
            self.assertEqual(expected, str(d))
            self.assertEqual(SOCKNAME, d.sock.getsockname())
            self.assertEqual(socket.AF_UNIX, d.sock.family)

    @unittest.skipUnless(hasattr(socket, "AF_UNIX") and sys.platform.startswith("linux"), "unix domain sockets required and linux os")
    def testDaemonUnixSocketAbstractNS(self):
        SOCKNAME = "\0test_unixsocket"  # mind the \0 at the start, for a Linux abstract namespace socket
        with Pyro4.core.Daemon(unixsocket=SOCKNAME) as d:
            locationstr = "./u:" + SOCKNAME
            self.assertEqual(locationstr, d.locationStr)
            self.assertEqual("PYRO:" + Pyro4.constants.DAEMON_NAME + "@" + locationstr, str(d.uriFor(Pyro4.constants.DAEMON_NAME)))
            # check the string representations
            expected = "<Pyro4.core.Daemon at 0x%x; %s - Unix; 1 objects>" % (id(d), locationstr)
            self.assertEqual(expected, str(d))
            sn_bytes = tobytes(SOCKNAME)
            self.assertEqual(sn_bytes, d.sock.getsockname())
            self.assertEqual(socket.AF_UNIX, d.sock.family)

    def testServertypeThread(self):
        old_servertype = config.SERVERTYPE
        config.SERVERTYPE = "thread"
        with Pyro4.core.Daemon(port=0) as d:
            self.assertIn(d.sock, d.sockets, "daemon's socketlist should contain the server socket")
            self.assertTrue(len(d.sockets) == 1, "daemon without connections should have just 1 socket")
        config.SERVERTYPE = old_servertype

    def testServertypeMultiplex(self):
        old_servertype = config.SERVERTYPE
        config.SERVERTYPE = "multiplex"
        with Pyro4.core.Daemon(port=0) as d:
            self.assertIn(d.sock, d.sockets, "daemon's socketlist should contain the server socket")
            self.assertTrue(len(d.sockets) == 1, "daemon without connections should have just 1 socket")
        config.SERVERTYPE = old_servertype

    def testServertypeFoobar(self):
        old_servertype = config.SERVERTYPE
        config.SERVERTYPE = "foobar"
        self.assertRaises(PyroError, Pyro4.core.Daemon)
        config.SERVERTYPE = old_servertype

    def testRegisterTwice(self):
        with Pyro4.core.Daemon(port=0) as d:
            o1 = MyObj("object1")
            d.register(o1)
            with self.assertRaises(DaemonError) as x:
                d.register(o1)
            self.assertEqual("object or class already has a Pyro id", str(x.exception))
            d.unregister(o1)
            d.register(o1, "samename")
            o2 = MyObj("object2")
            with self.assertRaises(DaemonError) as x:
                d.register(o2, "samename")
            self.assertEqual("an object or class is already registered with that id", str(x.exception))
            self.assertTrue(hasattr(o1, "_pyroId"))
            self.assertTrue(hasattr(o1, "_pyroDaemon"))
            d.unregister(o1)
            self.assertFalse(hasattr(o1, "_pyroId"))
            self.assertFalse(hasattr(o1, "_pyroDaemon"))
            o1._pyroId = "FOOBAR"
            with self.assertRaises(DaemonError) as x:
                d.register(o1)
            self.assertEqual("object or class already has a Pyro id", str(x.exception))
            o1._pyroId = ""
            d.register(o1)  # with empty-string _pyroId register should work

    def testRegisterTwiceForced(self):
        with Pyro4.core.Daemon(port=0) as d:
            o1 = MyObj("object1")
            d.register(o1, "name1")
            d.register(o1, "name2", force=True)
            d.register(o1, "name1", force=True)
            self.assertIs(d.objectsById["name1"], d.objectsById["name2"])
            d.unregister(o1)
            o1._pyroId = "FOOBAR_ID"
            d.register(o1, "newname", force=True)
            self.assertEqual("newname", o1._pyroId)
            self.assertIn("newname", d.objectsById)

    def testRegisterEtc(self):
        with Pyro4.core.Daemon(port=0) as d:
            self.assertEqual(1, len(d.objectsById))
            o1 = MyObj("object1")
            o2 = MyObj("object2")
            d.register(o1)
            self.assertRaises(DaemonError, d.register, o2, Pyro4.constants.DAEMON_NAME)  # cannot use daemon name
            d.register(o2, "obj2a")

            self.assertEqual(3, len(d.objectsById))
            self.assertEqual(o1, d.objectsById[o1._pyroId])
            self.assertEqual(o2, d.objectsById["obj2a"])
            self.assertEqual("obj2a", o2._pyroId)
            self.assertEqual(d, o2._pyroDaemon)

            # test unregister
            d.unregister("unexisting_thingie")
            self.assertRaises(ValueError, d.unregister, None)
            d.unregister("obj2a")
            d.unregister(o1._pyroId)
            self.assertEqual(1, len(d.objectsById))
            self.assertNotIn(o1._pyroId, d.objectsById)
            self.assertNotIn(o2._pyroId, d.objectsById)

            # test unregister objects
            del o2._pyroId
            d.register(o2)
            objectid = o2._pyroId
            self.assertIn(objectid, d.objectsById)
            self.assertEqual(2, len(d.objectsById))
            d.unregister(o2)
            # no more _pyro attributes must remain after unregistering
            for attr in vars(o2):
                self.assertFalse(attr.startswith("_pyro"))
            self.assertEqual(1, len(d.objectsById))
            self.assertNotIn(objectid, d.objectsById)
            self.assertRaises(DaemonError, d.unregister, [1, 2, 3])

            # test unregister daemon name
            d.unregister(Pyro4.constants.DAEMON_NAME)
            self.assertIn(Pyro4.constants.DAEMON_NAME, d.objectsById)

            # weird args
            w = MyObj("weird")
            self.assertRaises(AttributeError, d.register, None)
            self.assertRaises(AttributeError, d.register, 4444)
            self.assertRaises(TypeError, d.register, w, 666)

            # uri return value from register
            uri = d.register(MyObj("xyz"))
            self.assertIsInstance(uri, Pyro4.core.URI)
            uri = d.register(MyObj("xyz"), "test.register")
            self.assertEqual("test.register", uri.object)

    def testRegisterClass(self):
        with Pyro4.core.Daemon(port=0) as d:
            self.assertEqual(1, len(d.objectsById))
            d.register(MyObj)
            with self.assertRaises(DaemonError):
                d.register(MyObj)
            self.assertEqual(2, len(d.objectsById))
            d.uriFor(MyObj)
            # unregister:
            d.unregister(MyObj)
            self.assertEqual(1, len(d.objectsById))

    def testRegisterUnicode(self):
        with Pyro4.core.Daemon(port=0) as d:
            myobj1 = MyObj("hello1")
            myobj2 = MyObj("hello2")
            myobj3 = MyObj("hello3")
            uri1 = d.register(myobj1, "str_name")
            uri2 = d.register(myobj2, unicode("unicode_name"))
            uri3 = d.register(myobj3, "unicode_" + unichr(0x20ac))
            self.assertEqual(4, len(d.objectsById))
            uri = d.uriFor(myobj1)
            self.assertEqual(uri1, uri)
            _ = Pyro4.core.Proxy(uri)
            uri = d.uriFor(myobj2)
            self.assertEqual(uri2, uri)
            _ = Pyro4.core.Proxy(uri)
            uri = d.uriFor(myobj3)
            self.assertEqual(uri3, uri)
            _ = Pyro4.core.Proxy(uri)
            uri = d.uriFor("str_name")
            self.assertEqual(uri1, uri)
            _ = Pyro4.core.Proxy(uri)
            uri = d.uriFor(unicode("unicode_name"))
            self.assertEqual(uri2, uri)
            _ = Pyro4.core.Proxy(uri)
            uri = d.uriFor("unicode_" + unichr(0x20ac))
            self.assertEqual(uri3, uri)
            _ = Pyro4.core.Proxy(uri)

    def testDaemonObject(self):
        with Pyro4.core.Daemon(port=0) as d:
            daemon = Pyro4.core.DaemonObject(d)
            obj1 = MyObj("object1")
            obj2 = MyObj("object2")
            obj3 = MyObj("object2")
            d.register(obj1, "obj1")
            d.register(obj2, "obj2")
            d.register(obj3)
            daemon.ping()
            registered = daemon.registered()
            self.assertTrue(type(registered) is list)
            self.assertEqual(4, len(registered))
            self.assertIn("obj1", registered)
            self.assertIn("obj2", registered)
            self.assertIn(obj3._pyroId, registered)
            try:
                daemon.shutdown()
                self.fail("should not succeed to call unexposed method")
            except AttributeError:
                pass

    def testUriFor(self):
        d = Pyro4.core.Daemon(port=0)
        try:
            o1 = MyObj("object1")
            o2 = MyObj("object2")
            self.assertRaises(DaemonError, d.uriFor, o1)
            self.assertRaises(DaemonError, d.uriFor, o2)
            d.register(o1, None)
            d.register(o2, "object_two")
            o3 = MyObj("object3")
            self.assertRaises(DaemonError, d.uriFor, o3)  # can't get an uri for an unregistered object (note: unregistered name is ok)
            u1 = d.uriFor(o1)
            u2 = d.uriFor(o2._pyroId)
            u3 = d.uriFor("unexisting_thingie")  # unregistered name is no problem, it's just an uri we're requesting
            u4 = d.uriFor(o2)
            self.assertEqual(Pyro4.core.URI, type(u1))
            self.assertEqual("PYRO", u1.protocol)
            self.assertEqual("PYRO", u2.protocol)
            self.assertEqual("PYRO", u3.protocol)
            self.assertEqual("PYRO", u4.protocol)
            self.assertEqual("object_two", u4.object)
            self.assertEqual(Pyro4.core.URI("PYRO:unexisting_thingie@" + d.locationStr), u3)
        finally:
            d.close()

    def testDaemonWithStmt(self):
        d = Pyro4.core.Daemon()
        self.assertIsNotNone(d.transportServer)
        d.close()  # closes the transportserver and sets it to None
        self.assertIsNone(d.transportServer)
        with Pyro4.core.Daemon() as d:
            self.assertIsNotNone(d.transportServer)
            pass
        self.assertIsNone(d.transportServer)
        try:
            with Pyro4.core.Daemon() as d:
                print(1 // 0)  # cause an error
            self.fail("expected error")
        except ZeroDivisionError:
            pass
        self.assertIsNone(d.transportServer)
        d = Pyro4.core.Daemon()
        with d:
            pass
        try:
            with d:
                pass
            self.fail("expected error")
        except PyroError:
            # you cannot re-use a daemon object in multiple with statements
            pass
        d.close()

    def testRequestloopCondition(self):
        with Pyro4.core.Daemon(port=0) as d:
            condition = lambda: False
            start = time.time()
            d.requestLoop(loopCondition=condition)  # this should return almost immediately
            duration = time.time() - start
            self.assertLess(duration, 0.4)

    def testSimpleHandshake(self):
        conn = ConnectionMock()
        with Pyro4.core.Daemon(port=0) as d:
            self.sendHandshakeMessage(conn)
            success = d._handshake(conn)
            self.assertTrue(success)
            msg = Pyro4.message.Message.recv(conn, hmac_key=d._pyroHmacKey)
            self.assertEqual(Pyro4.message.MSG_CONNECTOK, msg.type)
            self.assertEqual(99, msg.seq)

    def testHandshakeDenied(self):
        class HandshakeFailDaemon(Pyro4.core.Daemon):
            def validateHandshake(self, conn, data):
                raise ValueError("handshake fail validation error")
        conn = ConnectionMock()
        with HandshakeFailDaemon(port=0) as d:
            self.sendHandshakeMessage(conn)
            success = d._handshake(conn)
            self.assertFalse(success)
            msg = Pyro4.message.Message.recv(conn, hmac_key=d._pyroHmacKey)
            self.assertEqual(Pyro4.message.MSG_CONNECTFAIL, msg.type)
            self.assertEqual(99, msg.seq)
            self.assertTrue(b"handshake fail validation error" in msg.data)
        with Pyro4.core.Daemon(port=0) as d:
            self.sendHandshakeMessage(conn)
            success = d._handshake(conn, denied_reason="no way, handshake denied")
            self.assertFalse(success)
            msg = Pyro4.message.Message.recv(conn, hmac_key=d._pyroHmacKey)
            self.assertEqual(Pyro4.message.MSG_CONNECTFAIL, msg.type)
            self.assertEqual(99, msg.seq)
            self.assertTrue(b"no way, handshake denied" in msg.data)

    def testCustomHandshake(self):
        conn = ConnectionMock()
        class CustomHandshakeDaemon(Pyro4.core.Daemon):
            def validateHandshake(self, conn, data):
                return ["sure", "have", "fun"]
            def annotations(self):
                return {"XYZZ": b"custom annotation set by daemon"}
        with CustomHandshakeDaemon(port=0) as d:
            corr_id = uuid.uuid4()
            self.sendHandshakeMessage(conn, correlation_id=corr_id)
            self.assertNotEqual(corr_id, current_context.correlation_id)
            success = d._handshake(conn)
            self.assertTrue(success)
            self.assertEqual(corr_id, current_context.correlation_id)
            msg = Pyro4.message.Message.recv(conn, hmac_key=d._pyroHmacKey)
            self.assertEqual(Pyro4.message.MSG_CONNECTOK, msg.type)
            self.assertEqual(99, msg.seq)
            self.assertEqual(2, len(msg.annotations))
            self.assertEqual(corr_id.bytes, msg.annotations["CORR"])
            self.assertEqual(b"custom annotation set by daemon", msg.annotations["XYZZ"])
            ser = Pyro4.util.get_serializer_by_id(msg.serializer_id)
            data = ser.deserializeData(msg.data, msg.flags & Pyro4.message.FLAGS_COMPRESSED)
            self.assertEqual(["sure", "have", "fun"], data)

    def testNAT(self):
        with Pyro4.core.Daemon() as d:
            self.assertIsNone(d.natLocationStr)
        with Pyro4.core.Daemon(nathost="nathosttest", natport=12345) as d:
            self.assertEqual("nathosttest:12345", d.natLocationStr)
            self.assertNotEqual(d.locationStr, d.natLocationStr)
            uri = d.register(MyObj(1))
            self.assertEqual("nathosttest:12345", uri.location)
            uri = d.uriFor("object")
            self.assertEqual("nathosttest:12345", uri.location)
            uri = d.uriFor("object", nat=False)
            self.assertNotEqual("nathosttest:12345", uri.location)
            d = Pyro4.core.Daemon(nathost="bla")
            self.assertTrue(d.natLocationStr.startswith("bla:"))
        try:
            _ = Pyro4.core.Daemon(natport=5555)
            self.fail("expected error")
        except ValueError:
            pass
        try:
            _ = Pyro4.core.Daemon(nathost="bla", natport=5555, unixsocket="testsock")
            self.fail("expected error")
        except ValueError:
            pass

    def testNATzeroPort(self):
        servertype = config.SERVERTYPE
        try:
            config.SERVERTYPE = "multiplex"
            with Pyro4.core.Daemon(nathost="nathosttest", natport=99999) as d:
                host, port = d.locationStr.split(":")
                self.assertNotEqual(99999, port)
                self.assertEqual("nathosttest:99999", d.natLocationStr)
            with Pyro4.core.Daemon(nathost="nathosttest", natport=0) as d:
                host, port = d.locationStr.split(":")
                self.assertEqual("nathosttest:%s" % port, d.natLocationStr)
            config.SERVERTYPE = "thread"
            with Pyro4.core.Daemon(nathost="nathosttest", natport=99999) as d:
                host, port = d.locationStr.split(":")
                self.assertNotEqual(99999, port)
                self.assertEqual("nathosttest:99999", d.natLocationStr)
            with Pyro4.core.Daemon(nathost="nathosttest", natport=0) as d:
                host, port = d.locationStr.split(":")
                self.assertEqual("nathosttest:%s" % port, d.natLocationStr)
        finally:
            config.SERVERTYPE = servertype

    def testNATconfig(self):
        try:
            config.NATHOST = None
            config.NATPORT = 0
            with Pyro4.core.Daemon() as d:
                self.assertIsNone(d.natLocationStr)
            config.NATHOST = "nathosttest"
            config.NATPORT = 12345
            with Pyro4.core.Daemon() as d:
                self.assertEqual("nathosttest:12345", d.natLocationStr)
        finally:
            config.NATHOST = None
            config.NATPORT = 0

    def testBehaviorDefaults(self):
        class TestClass:
            pass
        with Pyro4.core.Daemon() as d:
            d.register(TestClass)
            instance_mode, instance_creator = TestClass._pyroInstancing
            self.assertEqual("session", instance_mode)
            self.assertIsNone(instance_creator)

    def testInstanceCreationSingle(self):
        def creator(clazz):
            return clazz("testname")
        @Pyro4.core.behavior(instance_mode="single", instance_creator=creator)
        class TestClass:
            def __init__(self, name):
                self.name = name
        conn = Pyro4.socketutil.SocketConnection(socket.socket())
        d = Pyro4.core.Daemon()
        instance1 = d._getInstance(TestClass, conn)
        instance2 = d._getInstance(TestClass, conn)
        self.assertEqual("testname", instance1.name)
        self.assertIs(instance1, instance2)
        self.assertTrue(TestClass in d._pyroInstances)
        self.assertIs(instance1, d._pyroInstances[TestClass])
        self.assertFalse(TestClass in conn.pyroInstances)

    def testBehaviorDefaultsIsSession(self):
        class ClassWithDefaults:
            def __init__(self):
                self.name = "yep"
        conn1 = Pyro4.socketutil.SocketConnection(socket.socket())
        conn2 = Pyro4.socketutil.SocketConnection(socket.socket())
        d = Pyro4.core.Daemon()
        d.register(ClassWithDefaults)
        instance1a = d._getInstance(ClassWithDefaults, conn1)
        instance1b = d._getInstance(ClassWithDefaults, conn1)
        instance2a = d._getInstance(ClassWithDefaults, conn2)
        instance2b = d._getInstance(ClassWithDefaults, conn2)
        self.assertIs(instance1a, instance1b)
        self.assertIs(instance2a, instance2b)
        self.assertIsNot(instance1a, instance2a)
        self.assertFalse(ClassWithDefaults in d._pyroInstances)
        self.assertTrue(ClassWithDefaults in conn1.pyroInstances)
        self.assertTrue(ClassWithDefaults in conn2.pyroInstances)
        self.assertIs(instance1a, conn1.pyroInstances[ClassWithDefaults])
        self.assertIs(instance2a, conn2.pyroInstances[ClassWithDefaults])

    def testInstanceCreationSession(self):
        def creator(clazz):
            return clazz("testname")
        @Pyro4.core.behavior(instance_mode="session", instance_creator=creator)
        class ClassWithDecorator:
            def __init__(self, name):
                self.name = name
        conn1 = Pyro4.socketutil.SocketConnection(socket.socket())
        conn2 = Pyro4.socketutil.SocketConnection(socket.socket())
        d = Pyro4.core.Daemon()
        d.register(ClassWithDecorator)
        # check the class with the decorator first
        instance1a = d._getInstance(ClassWithDecorator, conn1)
        instance1b = d._getInstance(ClassWithDecorator, conn1)
        instance2a = d._getInstance(ClassWithDecorator, conn2)
        instance2b = d._getInstance(ClassWithDecorator, conn2)
        self.assertIs(instance1a, instance1b)
        self.assertIs(instance2a, instance2b)
        self.assertIsNot(instance1a, instance2a)
        self.assertFalse(ClassWithDecorator in d._pyroInstances)
        self.assertTrue(ClassWithDecorator in conn1.pyroInstances)
        self.assertTrue(ClassWithDecorator in conn2.pyroInstances)
        self.assertIs(instance1a, conn1.pyroInstances[ClassWithDecorator])
        self.assertIs(instance2a, conn2.pyroInstances[ClassWithDecorator])


    def testInstanceCreationPerCall(self):
        def creator(clazz):
            return clazz("testname")
        @Pyro4.core.behavior(instance_mode="percall", instance_creator=creator)
        class TestClass:
            def __init__(self, name):
                self.name = name
        with Pyro4.socketutil.SocketConnection(socket.socket()) as conn:
            with Pyro4.core.Daemon() as d:
                instance1 = d._getInstance(TestClass, conn)
                instance2 = d._getInstance(TestClass, conn)
                self.assertIsNot(instance1, instance2)
                self.assertFalse(TestClass in d._pyroInstances)
                self.assertFalse(TestClass in conn.pyroInstances)

    def testInstanceCreationWrongType(self):
        def creator(clazz):
            return Pyro4.core.URI("PYRO:test@localhost:9999")
        @Pyro4.core.behavior(instance_creator=creator)
        class TestClass:
            def method(self):
                pass
        with Pyro4.socketutil.SocketConnection(socket.socket()) as conn:
            with Pyro4.core.Daemon() as d:
                with self.assertRaises(TypeError):
                    d._getInstance(TestClass, conn)

    def testCombine(self):
        d1 = Pyro4.core.Daemon()
        d2 = Pyro4.core.Daemon()
        with self.assertRaises(TypeError):
            d1.combine(d2)
        d1.close()
        d2.close()
        try:
            config.SERVERTYPE = "multiplex"
            d1 = Pyro4.core.Daemon()
            d2 = Pyro4.core.Daemon()
            nsuri, nsd, bcd = Pyro4.naming.startNS(host="", bchost="")
            d1_selector = d1.transportServer.selector
            d1.combine(d2)
            d1.combine(nsd)
            d1.combine(bcd)
            self.assertIs(d1_selector, d1.transportServer.selector)
            self.assertIs(d1_selector, d2.transportServer.selector)
            self.assertIs(d1_selector, nsd.transportServer.selector)
            self.assertIs(d1_selector, bcd.transportServer.selector)
            self.assertEqual(4, len(d1.sockets))
            self.assertIn(d1.sock, d1.sockets)
            self.assertIn(d2.sock, d1.sockets)
            self.assertIn(nsd.sock, d1.sockets)
            self.assertIn(bcd, d1.sockets)
            bcd.close()
            nsd.close()
            d2.close()
            d1.close()
        finally:
            config.SERVERTYPE = "thread"


class MetaInfoTests(unittest.TestCase):
    def testMeta(self):
        with Pyro4.core.Daemon() as d:
            daemon_obj = d.objectsById[Pyro4.constants.DAEMON_NAME]
            self.assertTrue(len(daemon_obj.info()) > 10)
            meta = daemon_obj.get_metadata(Pyro4.constants.DAEMON_NAME)
            self.assertEqual({"get_metadata", "get_next_stream_item", "close_stream",
                              "info", "ping", "registered"}, meta["methods"])

    def testMetaSerialization(self):
        with Pyro4.core.Daemon() as d:
            daemon_obj = d.objectsById[Pyro4.constants.DAEMON_NAME]
            meta = daemon_obj.get_metadata(Pyro4.constants.DAEMON_NAME)
            for ser_id in [Pyro4.util.JsonSerializer.serializer_id,
                           Pyro4.util.MarshalSerializer.serializer_id,
                           Pyro4.util.PickleSerializer.serializer_id,
                           Pyro4.util.SerpentSerializer.serializer_id]:
                serializer = Pyro4.util.get_serializer_by_id(ser_id)
                data = serializer.dumps(meta)
                _ = serializer.loads(data)
            try:
                serializer = Pyro4.util.get_serializer_by_id(Pyro4.util.DillSerializer.serializer_id)
            except SerializeError:
                # dill doesn't work with ironpython so we allow an error here in that case
                if sys.platform != "cli":
                    raise
            else:
                data = serializer.dumps(meta)
                _ = serializer.loads(data)

    def testMetaResetCache(self):
        class Dummy:
            @Pyro4.core.expose
            def method(self):
                pass
        with Pyro4.core.Daemon() as d:
            dummy = Dummy()
            uri = d.register(dummy)
            daemon_obj = d.objectsById[Pyro4.constants.DAEMON_NAME]
            meta = daemon_obj.get_metadata(uri.object)
            self.assertNotIn("newly_added_method", meta["methods"])
            self.assertNotIn("newly_added_method_two", meta["methods"])
            Dummy.newly_added_method = Pyro4.core.expose(lambda self: None)
            meta = daemon_obj.get_metadata(uri.object)
            self.assertNotIn("newly_added_method", meta["methods"])
            d.resetMetadataCache(uri.object)
            meta = daemon_obj.get_metadata(uri.object)
            self.assertIn("newly_added_method", meta["methods"])
            Dummy.newly_added_method_two = Pyro4.core.expose(lambda self: None)
            d.resetMetadataCache(dummy)
            meta = daemon_obj.get_metadata(uri.object)
            self.assertIn("newly_added_method_two", meta["methods"])


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



import string
import time
from collections import Counter
from itertools import cycle
try:
    from itertools import izip_longest as zip_longest
except ImportError:
    from itertools import zip_longest
import Pyro4
import Pyro4.errors


@Pyro4.expose
class WordCounter(object):
    filter_words = {'a', 'an', 'at', 'the', 'i', 'he', 'she', 's', 'but', 'was', 'has', 'had', 'have', 'and',
                    'are', 'as', 'be', 'by', 'for', 'if', 'in', 'is', 'it', 'of', 'or', 'that',
                    'the', 'to', 'with', 'his', 'all', 'any', 'this', 'that', 'not', 'from', 'on',
                    'me', 'him', 'her', 'their', 'so', 'you', 'there', 'now', 'then', 'no', 'yes',
                    'one', 'were', 'they', 'them', 'which', 'what', 'when', 'who', 'how', 'where', 'some', 'my',
                    'into', 'up', 'out', 'some', 'we', 'us', 't', 'do'}
    trans_punc = {ord(punc): u' ' for punc in string.punctuation}

    def count(self, lines):
        counts = Counter()
        for num, line in enumerate(lines):
            if line:
                line = line.translate(self.trans_punc).lower()
                interesting_words = [w for w in line.split() if w.isalpha() and w not in self.filter_words]
                counts.update(interesting_words)
            if num % 10 == 0:
                time.sleep(0.01)  # artificial delay to show execution time differences (and make this not cpu-bound)
        return counts


def grouper(n, iterable, padvalue=None):
    """grouper(3, 'abcdefg', 'x') --> ('a','b','c'), ('d','e','f'), ('g','x','x')"""
    return zip_longest(*[iter(iterable)]*n, fillvalue=padvalue)


@Pyro4.expose
class Dispatcher(object):
    def count(self, lines):
        # use the name server's prefix lookup to get all registered wordcounters
        with Pyro4.locateNS() as ns:
            all_counters = ns.list(prefix="example.dc.wordcount.")
        counters = [Pyro4.Proxy(uri) for uri in all_counters.values()]
        for c in counters:
            c._pyroAsync()   # set proxy in asynchronous mode
        roundrobin_counters = cycle(counters)

        # chop the text into chunks that can be distributed across the workers
        # uses asynchronous proxy so that we can hand off everything in parallel
        # counter is selected in a round-robin fashion from list of all available counters
        # (This is a brain dead way because it doesn't scale - all the asynchronous calls are hogging
        # the worker threads in the server. That's why we've increased that a lot at the start
        # of this file, just for the sake of this example!)
        async_results = []
        for chunk in grouper(100, lines):
            counter = next(roundrobin_counters)
            result = counter.count(chunk)
            async_results.append(result)

        # gather the results
        print("Collecting %d results..." % len(async_results))
        totals = Counter()
        for result in async_results:
            try:
                totals.update(result.value)
            except Pyro4.errors.CommunicationError as x:
                raise Pyro4.errors.PyroError("Something went wrong in the server when collecting the async responses: "+str(x))
        for proxy in counters:
            proxy._pyroRelease()
        return totals


if __name__ == "__main__":
    print("Spinning up 5 wordcounters, and 1 dispatcher.")
    Pyro4.config.SERVERTYPE = "thread"
    Pyro4.Daemon.serveSimple(
        {
            WordCounter(): "example.dc.wordcount.1",
            WordCounter(): "example.dc.wordcount.2",
            WordCounter(): "example.dc.wordcount.3",
            WordCounter(): "example.dc.wordcount.4",
            WordCounter(): "example.dc.wordcount.5",
            Dispatcher:    "example.dc.dispatcher"
        }, verbose=False
    )



from __future__ import print_function
import random
import threading
import time


class StockMarket(object):
    def __init__(self, marketname, symbols):
        self.name = marketname
        self.symbolmeans = {}
        for symbol in symbols:
            self.symbolmeans[symbol] = random.uniform(20, 200)
        self.aggregators = []

    def generate(self):
        quotes = {}
        for symbol, mean in self.symbolmeans.items():
            if random.random() < 0.2:
                quotes[symbol] = round(random.normalvariate(mean, 20), 2)
        print("new quotes generated for", self.name)
        for aggregator in self.aggregators:
            aggregator.quotes(self.name, quotes)

    def listener(self, aggregator):
        print("market {0} adding new aggregator".format(self.name))
        self.aggregators.append(aggregator)

    def symbols(self):
        return self.symbolmeans.keys()

    def run(self):
        def generate_symbols():
            while True:
                time.sleep(random.random())
                self.generate()

        thread = threading.Thread(target=generate_symbols)
        thread.setDaemon(True)
        thread.start()


def main():
    nasdaq = StockMarket("NASDAQ", ["AAPL", "CSCO", "MSFT", "GOOG"])
    newyork = StockMarket("NYSE", ["IBM", "HPQ", "BP"])
    nasdaq.run()
    newyork.run()
    return [nasdaq, newyork]



# this example uses a background thread to run the daemon in, also works on Windows

from __future__ import print_function
import socket
import threading
import Pyro4

# it's okay to use Pickle because we trust our own processes, and it is efficient
Pyro4.config.SERIALIZER = "pickle"
Pyro4.config.SERIALIZERS_ACCEPTED |= {"pickle"}


# create our own socket pair (server-client sockets that are already connected)
sock1, sock2 = socket.socketpair()


class Echo(object):
    @Pyro4.expose
    def echo(self, message):
        print("server got message: ", message)
        return "thank you"


# create a daemon with some Pyro objectrunning on our custom server socket
daemon = Pyro4.Daemon(connected_socket=sock1)
daemon.register(Echo, "echo")
print("(Pyro daemon running on", daemon.locationStr, ")\n")
daemonthread = threading.Thread(target=daemon.requestLoop)
daemonthread.daemon = True
daemonthread.start()


# create a client running on the client socket
with Pyro4.Proxy("echo", connected_socket=sock2) as p:
    reply = p.echo("hello!")
    print("client got reply:", reply)
    reply = p.echo("hello again!")
    print("client got reply:", reply)
with Pyro4.Proxy("echo", connected_socket=sock2) as p:
    reply = p.echo("hello2!")
    print("client got reply:", reply)
    reply = p.echo("hello2 again!")
    print("client got reply:", reply)

print("\nThe end.")
